/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * Epipolar.hpp
 *
 *  Created on: 2010/11/15
 *      Author: HARK Development Team
 */

#ifndef EPIPOLAR_HPP_
#define EPIPOLAR_HPP_

#include <DimensionPromise/types.hpp>
#include <AngleConstants.hpp>

namespace NSHarkBinaural
{

/**
 * 信号源からの距離差を計算するクラス.
 */
class Epipolar
{
public:
    /**
     * Epipolarクラスを生成する.
     * @param sourceDistance 信号源から頭部の中心までの距離
     * @param headRadius 頭部の半径
     */
    Epipolar(NSDimensionPromise::Length sourceDistance, NSDimensionPromise::Length headRadius)
        : sourceDistance(sourceDistance), headRadius(headRadius)
    { }

    /**
     * 右チャンネルに対する左チャンネルの信号源からの距離差を求める.
     * 従来の epipolar geometry の手法を用いて、
     * 右チャンネルに対する左チャンネルの信号源からの距離差を求める。
     * @param sourceDirection 信号源の方向 中心より左を正、中心より右を負とする
     * @return 距離差 左の方が距離が短い場合は負、右の方が距離が短い場合は正
     */
    NSDimensionPromise::Length LRDifferenceSimple(NSDimensionPromise::Angle sourceDirection) const;

    /**
     * 右チャンネルに対する左チャンネルの信号源からの距離差を求める.
     * 回り込みを考慮に入れた epipolar geometry の手法を用いて、
     * 右チャンネルに対する左チャンネルの信号源からの距離差を求める。
     * @param sourceDirection 信号源の方向 中心より左を正、中心より右を負とする
     * @return 距離差 左の方が距離が短い場合は負、右の方が距離が短い場合は正
     */
    NSDimensionPromise::Length LRDifferenceRevised(NSDimensionPromise::Angle sourceDirection) const;

private:
    /**
     * 信号源から頭部の中心までの距離.
     */
    NSDimensionPromise::Length sourceDistance;

    /**
     * 頭部の半径.
     */
    NSDimensionPromise::Length headRadius;
};

/**
 * 指定された周波数の音が指定された距離だけ進むときの位相差を計算する.
 * @param lengthDiff 音が進む距離
 * @param signalFreq 音の周波数
 * @param soundSpeed 1秒間に音が進む距離（音速）
 */
template <typename T>
const NSDimensionPromise::AngleBase<T>
PhaseDifference(
    const NSDimensionPromise::LengthBase<T>& lengthDiff,
    const NSDimensionPromise::FrequencyBase<T> signalFreq,
    const NSDimensionPromise::Velocity1Base<T>& soundSpeed = NSDimensionPromise::Velocity1Base<T>(340.0))
{
    return NSDimensionPromise::AngleBase<T>(
               lengthDiff / soundSpeed * 2.0 * Angles::AnglePi.ToRadian() * signalFreq);
}

}

#endif /* EPIPOLAR_HPP_ */
