/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * InterauralDifference.hpp
 *
 *  Created on: 2010/11/11
 *      Author: HARK Development Team
 */

/**
 * @file InterauralDifference.hpp
 * 両耳間の差異を扱うクラスなど.
 * IID(Interaural Intensity Difference)、IPD(Interaural Phase Difference)
 * などを扱うクラスを集めたファイル
 */

#ifndef INTERAURALDIFFERENCE_HPP_
#define INTERAURALDIFFERENCE_HPP_

#include <valarray>
#include <stdexcept>
#include <stdint.h>
#include <boost/shared_ptr.hpp>
#include <Channel.hpp>
#include <Signal.hpp>
#include <Utils.hpp>
#include <Logger.hpp>

namespace FD
{
template <typename T> class RCPtr;
template <typename T> class Vector;
}

namespace NSHarkBinaural
{

/**
 * 両耳間の差異を表すベースクラス.
 * @tparam T 差異の1要素を表す型
 */
template <typename T>
class InterauralDifference
{
public:
    /**
     * 左右両チャンネルのスペクトルを表す型.
     */
    typedef ChannelHolder<Spectrum::ptr_t, Spectrum::ptr_t>::ptr_t LRSpectraPointer;

    /**
     * スマートポインタ型
     */
    typedef boost::shared_ptr<InterauralDifference> ptr_t;

    /**
     * 強度差配列の各要素の型
     */
    typedef T ElementType;

    /**
     * クラスを破棄する.
     */
    virtual ~InterauralDifference()
    {
        logger.Finest(boost::format("InterauralDifference::~InterauralDifference %08x") %
                      reinterpret_cast<uintptr_t>(this));
    }

#if 0
    /**
     * 差異を取得する.
     * @return 差異配列
     */
    std::valarray<T>& Difference()
    {
        return this->difference;
    }

    /**
     * 差異を取得する.
     * @return 差異配列
     */
    const std::valarray<T>& Difference() const
    {
        return this->difference;
    }
#else
    // valarrayへの参照を返すのではなく、valarrayの各メソッドを呼び出すような設計にする。
    // そうすれば、さまざまな変更に対して柔軟になる。
    // 例えば、「valarrayの各要素アクセスのログを取りたい」「遅延評価方式にしたい」とか。

    /**
     * 指定されたインデックスにある要素を得る.
     * @param index 要素のインデックス
     * @return 要素への参照
     */
    T& operator [](size_t index)
    {
        return this->difference[index];
    }

    /**
     * 指定されたインデックスにある要素を得る.
     * @param index 要素のインデックス
     * @return 要素への参照
     */
    const T& operator [](size_t index) const
    {
        return this->difference[index];
    }

    /**
     * 指定されたインデックスにある要素を得る.
     * @param index 要素のインデックス
     * @return 要素への参照
     */
    T& At(size_t index)
    {
        RangeCheck(index);
        return this->difference[index];
    }
    /**
     * 指定されたインデックスにある要素を得る.
     * @param index 要素のインデックス
     * @return 要素への参照
     */
    const T& At(size_t index) const
    {
        RangeCheck(index);
        return this->difference[index];
    }

    /**
     * 強度差配列の要素数を得る.
     * @return 要素数
     */
    size_t Size() const
    {
        return this->difference.size();
    }

    // 算術代入演算子は定義していない。使わなそうだからサボっているだけ。

#endif

    /**
     * 計算の基準側を取得する.
     * @return 計算の基準
     */
    Channel::Kind RelativeTo() const
    {
        return this->relativeTo;
    }

protected:
    /**
     * デフォルトコンストラクタ.
     * このクラスのインスタンス化を禁止するためにprotectedにしている。
     */
    InterauralDifference(Channel::Kind relativeTo = Channel::L)
        : difference(), relativeTo(relativeTo), logger(Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME))
    {
        logger.Finest(boost::format("InterauralDifference::InterauralDifference %08x") %
                      reinterpret_cast<uintptr_t>(this));
    }

    /**
     * 差異配列を得る.
     * @return 差異配列
     */
    std::valarray<T>& Difference()
    {
        return this->difference;
    }

    /**
     * 差異配列を得る.
     * @return 差異配列
     */
    const std::valarray<T>& Difference() const
    {
        return this->difference;
    }

private:
    /**
     * 指定されたインデックスが配列の範囲にあるかどうか.
     */
    void RangeCheck(size_t index)
    {
        if (index >= this->difference.size())
        {
            throw std::out_of_range("index out of range");
        }
    }

    /**
     * relativeTo側を基準とした、左右の強度差配列
     */
    std::valarray<T> difference;

    /**
     * 差異の基準側
     */
    Channel::Kind relativeTo;

    Logger& logger;
};

//template <typename T>
//Logger& InterauralDifference<T>::logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

/**
 * 両耳間強度差を表すクラス.
 */
class InterauralIntensityDifference : public InterauralDifference<Spectrum::ElementType::value_type>
{
public:
    /**
     * スマートポインタ型
     */
    typedef boost::shared_ptr<InterauralIntensityDifference> ptr_t;

    /**
     * 両耳のスペクトルから差異を計算しクラスを生成する.
     * 引数で指定された両耳のスペクトルを用いて、両耳間強度差を計算し、クラスを生成する
     * @param input 両耳のスペクトル
     * @param relativeTo 差異を計算するときに基準とする側
     */
    InterauralIntensityDifference(LRSpectraPointer input, Channel::Kind relativeTo);

    /**
     * クラスを破棄する.
     */
    virtual ~InterauralIntensityDifference()
    {
        logger.Finest("InterauralPhaseDifference::~InterauralPhaseDifference");
    }

private:

    Logger& logger;
};


/**
 * 両耳間強度差の別名
 */
typedef InterauralIntensityDifference IID;

/**
 * InterauralIntensityDifferenceからFD::Vectorへ変換する.
 * @param iid 変換元のiid
 * @return 変換された配列
 */
FD::RCPtr<FD::Vector<IID::ElementType> > ToFDVector(IID::ptr_t iid);


/**
 * 両耳間位相差を表すクラス.
 */
class InterauralPhaseDifference : public InterauralDifference<Spectrum::ElementType::value_type>
{
public:
    /**
     * スマートポインタ型
     */
    typedef boost::shared_ptr<InterauralPhaseDifference> ptr_t;

    /**
     * 両耳のスペクトルから差異を計算しクラスを生成する.
     * 引数で指定された両耳のスペクトルを用いて、両耳間位相差を計算し、クラスを生成する
     * @param input 両耳のスペクトル
     * @param relativeTo 差異を計算するときに基準とする側
     */
    InterauralPhaseDifference(LRSpectraPointer input, Channel::Kind relativeTo);

    /**
     * クラスを破棄する.
     */
    virtual ~InterauralPhaseDifference()
    {
        logger.Finest("InterauralPhaseDifference::~InterauralPhaseDifference");
    }

private:

    Logger& logger;
};

/**
 * 両耳間位相差の別名
 */
typedef InterauralPhaseDifference IPD;

/**
 * InterauralPhaseDifferenceからFD::Vectorへ変換する.
 * @param ipd 変換元のipd
 * @return 変換された配列
 */
FD::RCPtr<FD::Vector<IPD::ElementType> > ToFDVector(IPD::ptr_t ipd);

}

#endif /* INTERAURALDIFFERENCE_HPP_ */
