/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * DirectionPassFilter.cpp
 *
 *  Created on: 2010/11/15
 *      Author: HARK Development Team
 */

#include <algorithm>

#define NDEBUG

#include <DimensionPromise/dimpro.hpp>

#include <DirectionPassFilter.hpp>
#include <MaxMin.hpp>
#include <constants.hpp>
#include <Debug.hpp>
#include <Epipolar.hpp>
#include <LRSignal.hpp>
#include <Logger.hpp>
#include <IIPDS.hpp>

using namespace std;
using namespace boost;
using namespace NSDimensionPromise;

namespace NSHarkBinaural
{

void CalculateMaxMin(
    NSDimensionPromise::Angle theta,
    NSDimensionPromise::Angle width,
    MaxMin<NSDimensionPromise::Angle>& mma/*, MaxMin<int>& mmi*/)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    LogFine(logger, "CalculateMaxMin");

    bool upper = false, lower = false;

    const Angle maxa(theta + width);
    const Angle mina(theta - width);

    // +PI/2の範囲を越えるかどうか
    if (mina.ToRadian() < Constants::ConstPi2 && maxa.ToRadian() > Constants::ConstPi2)
    {
        upper = true;
    }

    // -PI/2の範囲を越えるかどうか
    if (mina.ToRadian() < -Constants::ConstPi2 && maxa.ToRadian() > -Constants::ConstPi2)
    {
        lower = true;
    }

    if (upper && lower)
    {
        // 両方とも越えていれば、上限下限を +PI/2, -PI/2 にする
        mma.max = Angle(Constants::ConstPi2);
        mma.min = Angle(-Constants::ConstPi2);
    }
    else if (upper && !lower)
    {
        // +PI/2の範囲を越えるが、-PI/2の範囲を越えていない場合
        mma.max = Angle(Constants::ConstPi2);

        //if (fabs(mina.ToRadian() - Constants::M_PI_2) > fabs(Constants::M_PI_2 - maxa.ToRadian()))
        if (theta.ToRadian() < Constants::ConstPi2)
        {
            mma.min = mina;
        }
        else
        {
            mma.min = Angle(Constants::ConstPi) - maxa;
        }
    }
    else if (!upper && lower)
    {
        // -PI/2の範囲を越えるが、+PI/2の範囲を越えていない場合
        mma.min = Angle(-Constants::ConstPi2);

        //if (fabs(-Constants::M_PI_2 - mina.ToRadian()) > fabs(maxa.ToRadian() + Constants::M_PI_2))
        if (theta.ToRadian() < -Constants::ConstPi2)
        {
            mma.max = Angle(-Constants::ConstPi) - mina;
        }
        else
        {
            mma.max = maxa;
        }
    }
    else
    {
        mma.max = maxa;
        mma.min = mina;
    }

#if 0
    mmi.min = static_cast<int>( (mma.min.ToDegree() + 90) / 5 + 0.5 );
    mmi.max = static_cast<int>( (mma.max.ToDegree() + 90) / 5 + 0.5 );

    if (mmi.max > 36)
    {
        DebugPuts(" mmi.max > 36");
        mmi.max = 36;
    }
    if (mmi.min < 0)
    {
        DebugPuts(" mmi.min < 0");
        mmi.min = 0;
    }
#endif
}

void PassBandFlag::Flag(PassBandSetMode::Kind setMode, Kind flag)
{
    switch (flag)
    {
    case PassBandFlag::CutOff: // 遮断フラグを設定する
        switch (setMode)
        {
        case PassBandSetMode::CutOffOnly:
        case PassBandSetMode::CutOffAndPass:
            this->flag = flag;
            break;
        default:
            break;
        }
        break;
    case PassBandFlag::Pass: // 通過フラグを設定する
        switch (setMode)
        {
        case PassBandSetMode::PassOnly:
        case PassBandSetMode::CutOffAndPass:
            this->flag = flag;
            break;
        default:
            break;
        }
        break;
    default: // フラグの値が知らない値だった
        break;
    }
}

void PassBand::SetPassBandEpipolar(
    PassBandSetMode::Kind setMode,
    SeparationSetting::ptr_t separationSetting,
    NSDimensionPromise::Angle dir, NSDimensionPromise::Angle range,
    IPD::ptr_t ipd,
    Unspectrum::ptr_t signalPowerSpectrum, Unspectrum::ptr_t roomnoisePowerSpectrum,
    NSDimensionPromise::Length sourceDistance, NSDimensionPromise::Length headRadius)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    MaxMin<Angle> mm, ipdOfIndex;

    CalculateMaxMin(dir, range, mm);

    Epipolar epipolar(sourceDistance, headRadius);

    // 音路差下限
    Length diffMin = epipolar.LRDifferenceRevised(mm.min);

    // 音路差上限
    Length diffMax = epipolar.LRDifferenceRevised(mm.max);

    LogFine(logger, format(" SPBE %f - %f - %f") % Angle(dir - range).ToDegree() % dir.ToDegree() % Angle(dir + range).ToDegree());

    Frequency freq;
    int f1500 = (int)(Frequency(1500) / separationSetting->samplingFrequency * separationSetting->ndft).X();

    format msg1("%f Hz Fs %d NDFT %d (ipdl:%f <= ipd:%f <= ipdh:%f)");

    for (size_t i = 0; i < ipd->Size(); ++i)
    {
        // i番目の周波数帯の下限周波数を求める
        // i番目の周波数帯は freq ～ freq + サンプリング周波数 / DFTサンプル数
        freq = static_cast<double>(i) * separationSetting->samplingFrequency / static_cast<double>(separationSetting->ndft);
        ipdOfIndex.min = PhaseDifference( min(diffMax, diffMin), freq );
        ipdOfIndex.max = PhaseDifference( max(diffMax, diffMin), freq );

        LogFine(logger, msg1 %
                freq.ToHz() % (int)separationSetting->samplingFrequency.ToHz() % separationSetting->ndft %
                ipdOfIndex.min.ToRadian() % ipd->At(i) % ipdOfIndex.max.ToRadian());

        if ( ipdOfIndex.min > ipdOfIndex.max )
        {
            LogWarning(logger, "PassBand::SetPassBandEpipolar: ERROR: ipdl > ipdh");
            //cerr << "PassBand::SetPassBandEpipolar: ERROR: ipdl > ipdh" << endl;
        }

        if ( ipdOfIndex.min.ToRadian() < ipd->At(i) && ipd->At(i) < ipdOfIndex.max.ToRadian() &&
                i < f1500)
        {
            // 該当する方向であれば、通過に設定する
            this->passBand[i].Flag(setMode, PassBandFlag::Pass);
        }
        else
        {
            // 該当する方向でなければ、遮断に設定する
            this->passBand[i].Flag(setMode, PassBandFlag::CutOff);
        }
    }
}

void PassBand::SetPassBandHRTF(
    PassBandSetMode::Kind setMode,
    SeparationSetting::ptr_t separationSetting,
    NSDimensionPromise::Angle dir, NSDimensionPromise::Angle range,
    IPD::ptr_t ipd, IID::ptr_t iid,
    LRUnspectra::ptr_t signalPowerSpectra, LRUnspectra::ptr_t roomnoisePowerSpectra,
    IIPDS& iipds,
    Unspectrum::ElementType thresDB)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    MaxMin<Angle> mm;
    MaxMin<int> mmi;
    MaxMin<double> mmipd, mmiid;
    CalculateMaxMin(dir, range, mm);
    mmi.min = static_cast<int>( (mm.min.ToDegree() + 90) / iipds.dirtick + 0.5 );
    mmi.max = static_cast<int>( (mm.max.ToDegree() + 90) / iipds.dirtick + 0.5 );

    size_t f1500 = static_cast<size_t>((Frequency(1500) / separationSetting->samplingFrequency * separationSetting->ndft).X());

    if (mmi.min < 0 || iipds.ndir <= mmi.max)
    {
        LogSevere(logger, format("mmi is invalid, mm.min=%fdeg, mm.max=%fdeg, mmi=%d,%d, iipds.ndir=%d") % mm.min.ToDegree() % mm.max.ToDegree() % mmi.min % mmi.max % iipds.ndir);
        throw logic_error("mmi is invalid");
    }

    for (size_t i = 0; i < separationSetting->ndft / 2 + 1; ++i)
    {
        // 与えられた角度の範囲で、ipd, iidの最大、最小を探す
        mmipd.min = mmipd.max = iipds.ipd[mmi.min].m[i];
        mmiid.min = mmiid.max = iipds.iid[mmi.min].m[i];
        for (int j = mmi.min + 1; j <= mmi.max; ++j)
        {
            mmipd.min = min(mmipd.min, iipds.ipd[j].m[i]);
            mmipd.max = max(mmipd.max, iipds.ipd[j].m[i]);
            mmiid.min = min(mmiid.min, iipds.iid[j].m[i]);
            mmiid.max = max(mmiid.max, iipds.iid[j].m[i]);
        }

        if (((i <= f1500 && ipd->At(i) > mmipd.min && ipd->At(i) < mmipd.max) ||
                (i > f1500 && iid->At(i) > mmiid.min && iid->At(i) < mmiid.max)) &&
                signalPowerSpectra->Left()->At(i) > roomnoisePowerSpectra->Left()->At(i) + thresDB &&
                signalPowerSpectra->Right()->At(i) > roomnoisePowerSpectra->Right()->At(i) + thresDB)
        {
            // 該当する方向であれば、通過に設定する
            this->passBand[i].Flag(setMode, PassBandFlag::Pass);
        }
        else
        {
            // 該当する方向でなければ、遮断に設定する
            this->passBand[i].Flag(setMode, PassBandFlag::CutOff);
        }
    }
}

void SetPassBandFlags(PassBand::ptr_t passBand, PassBandFlag::Kind flag)
{
    for (size_t i = 0; i < passBand->Size(); ++i)
    {
        passBand->At(i).Flag(PassBandSetMode::CutOffAndPass, flag);
    }
}

BandExtractor::LRSpectra::ptr_t BandExtractor::Extract(PassBand::ptr_t passBand)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    LRSpectra::ptr_t result;
    // スペクトルの長さと、通過帯情報の長さを調べる
    if (passBand.get() != 0 && IsEqualSignalSize(this->lrspectra) &&
            this->lrspectra->Left()->Size() == passBand->Size())
    {
        // 長さが等しいなら、処理する

        size_t size = passBand->Size();

        // 使いやすいように、入力スペクトルを取り出す
        Spectrum::ptr_t lIn = this->lrspectra->Left();
        Spectrum::ptr_t rIn = this->lrspectra->Left();

        // 出力スペクトルを生成する
        Spectrum::ptr_t lOut(new Spectrum(size));
        Spectrum::ptr_t rOut(new Spectrum(size));
        // 出力変数に設定
        result = LRSpectra::ptr_t(new LRSpectra(lOut, rOut));

        // 通過帯情報にしたがい、抽出する
        for (size_t i = 0; i < passBand->Size(); ++i)
        {
            if (passBand->At(i).Flag() == PassBandFlag::Pass)
            {
                // passBandで通過する部分は入力から出力へコピーする
                lOut->At(i) = lIn->At(i);
                rOut->At(i) = rIn->At(i);
                LogFine(logger, " *** PASS *** copying Spectrum");
            }
            else if (passBand->At(i).Flag() == PassBandFlag::CutOff)
            {
#ifndef RESIDUE
                // passBandで遮断される部分は0にする
                lOut->At(i) = Spectrum::ElementType();
                rOut->At(i) = Spectrum::ElementType();
#else
                // passBandで遮断される部分にもコピーする
                lOut->At(i) = lIn->At(i);
                rOut->At(i) = rIn->At(i);
#endif
                LogFine(logger, " ***CUTOFF*** copying ZERO");
            }
        }
    }
    return result;
}

}
